#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from .XMLTools import XMLTools
from .SVGStyle import SVGStyle
from .Vector2D import Vector2D

class SVGXYObject():
	_X_ATTRIBUTE_NAME = "x"
	_Y_ATTRIBUTE_NAME = "y"

	@property
	def pos(self):
		return Vector2D(x = self._get_float_attribute(self._X_ATTRIBUTE_NAME), y = self._get_float_attribute(self._Y_ATTRIBUTE_NAME))

	@pos.setter
	def pos(self, value: Vector2D):
		self.node.setAttribute(self._X_ATTRIBUTE_NAME, str(value.x))
		self.node.setAttribute(self._Y_ATTRIBUTE_NAME, str(value.y))

class SVGWidthHeightObject():
	_DEFAULT_WIDTH = 0
	_DEFAULT_HEIGHT = 0

	def _deunify(self, name, default_value):
		value = self._default_get_attribute(name)
		if value is None:
			return default_value
		if value.endswith("mm"):
			return float(value[:-2]) / 25.4 * 96
		else:
			return float(value)

	@property
	def extents(self):
		return Vector2D(x = self._deunify("width", self._DEFAULT_WIDTH), y = self._deunify("height", self._DEFAULT_HEIGHT))

	@extents.setter
	def extents(self, value: Vector2D):
		self.node.setAttribute("width", str(value.x))
		self.node.setAttribute("height", str(value.y))


class SVGStyleObject():
	@property
	def style(self):
		return SVGStyle.from_node(self.node, auto_sync = True)


class SVGObject():
	_TAG_NAME = None
	_REGISTERED_CLASSES = { }

	def __init__(self, node):
		self._node = node

	@property
	def svg_document(self):
		doc = self.node.ownerDocument
		svg_doc = getattr(doc, "_pysvgedit", None)
		return svg_doc

	@property
	def node(self):
		return self._node

	@property
	def svgid(self):
		return self._default_get_attribute("id")

	@svgid.setter
	def svgid(self, value):
		self.node.setAttribute("id", value)

	@property
	def label(self):
		return self._default_get_attribute("inkscape:label")

	@label.setter
	def label(self, value: str):
		self.node.setAttribute("inkscape:label", value)

	def _default_get_attribute(self, name, default_value = None):
		return XMLTools.default_get_attribute(self.node, name, default_value = default_value)

	def _get_float_attribute(self, name):
		return float(self._default_get_attribute(name, default_value = 0))

	@classmethod
	def _new_element(cls):
		return XMLTools.new_element(cls.get_tagname())

	def add(self, svg_object):
		self.node.appendChild(svg_object.node)
		svg_object.node.ownerDocument = self.node.ownerDocument
		if svg_object.svgid is None:
			svg_object.svgid = self.svg_document.get_unused_id()
		if hasattr(svg_object, "post_add_hook"):
			svg_object.post_add_hook(self)
			svg_object.post_add_hook = None
		return svg_object

	@classmethod
	def get_tagname(cls):
		assert(cls._TAG_NAME is not None)
		return cls._TAG_NAME

	def get(self, object_class, constraint = None):
		object_class = self._resolve_object_class(object_class)
		for child in XMLTools.find_all_elements(self.node, object_class.get_tagname()):
			child = object_class(child)
			if (constraint is None) or constraint(child):
				yield child

	def get_first(self, object_class, constraint = None):
		return next(self.get(object_class, constraint = constraint))

	def walk(self, object_class, constraint = None):
		object_class = self._resolve_object_class(object_class)
		for node in XMLTools.walk_elements(self.node, object_class.get_tagname()):
			node = object_class(node)
			if (constraint is None) or constraint(node):
				yield node

	def _resolve_object_class(self, object_class):
		if isinstance(object_class, str):
			if object_class in self._REGISTERED_CLASSES:
				return self._REGISTERED_CLASSES[object_class]
			else:
				raise ValueError(f"Class named '{object_class}' does not have a registered handler.")
		else:
			return object_class

	@classmethod
	def register(cls, svg_object_class):
		if svg_object_class._TAG_NAME in cls._REGISTERED_CLASSES:
			raise ValueError(f"{svg_object_class._TAG_NAME} of {svg_object_class} already registered.")
		cls._REGISTERED_CLASSES[svg_object_class._TAG_NAME] = svg_object_class
		return svg_object_class



#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os
import sys
import tempfile
import subprocess
import pysvgedit
from .FriendlyArgumentParser import FriendlyArgumentParser

class AnimationRendererApp():
	def __init__(self, args):
		self._args = args

	@property
	def outdir(self):
		if self._args.outdir is None:
			return ""
		else:
			return self._args.outdir + "/"

	def run(self):
		self._doc = pysvgedit.SVGDocument.readfile(self._args.infile_svg)
		self._anim = pysvgedit.SVGAnimation(self._doc, animation_mode = pysvgedit.SVGAnimationMode(self._args.animation_mode))
		tvars = {
			"prefix": os.path.splitext(self._args.infile_svg)[0],
		}
		for (frameno, frame) in enumerate(self._anim, 1):
			tvars["frameno"] = frameno
			output_filename = f"{self.outdir}{self._args.filename_template.format(**tvars)}"
			if self._args.verbose >= 1:
				print(f"Frame {frameno}: {output_filename}")
			if not self._args.inkscape_render:
				frame.writefile(output_filename)
			else:
				with tempfile.NamedTemporaryFile(prefix = "pysvgedit_", suffix = ".svg", mode = "w") as f:
					frame.write(f)
					f.flush()
					subprocess.check_call([ "inkscape", "-o", output_filename, f.name ])

	@classmethod
	def main(cls):
		parser = FriendlyArgumentParser(description = "Render an animated SVG file.")
		parser.add_argument("-i", "--inkscape-render", action = "store_true", help = "Do not output raw SVG data, but render through Inkscape. Allows direct generation of PDF or PNG output.")
		parser.add_argument("-m", "--animation-mode", choices = [ "compose", "compose-all", "replace" ], default = "compose", help = "Specify the animation mode to render in. Can be one of %(choices)s, defaults to %(default)s.")
		parser.add_argument("-n", "--filename-template", metavar = "template", default = "{prefix}_{frameno:02d}.svg", help = "Can specify a filename template. Defaults to '%(default)s'.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("infile_svg", help = "Input SVG file.")
		parser.add_argument("outdir", nargs = "?", help = "Output directory to render frames in. Defaults to current directory if omitted.")
		args = parser.parse_args(sys.argv[1:])

		app = cls(args)
		return app.run()



#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import re
import contextlib
import dataclasses
from .SVGObject import SVGObject, SVGStyleObject
from .Vector2D import Vector2D

@dataclasses.dataclass
class SVGPathElementClose():
	def apply(self, pos):
		return pos

@dataclasses.dataclass
class SVGPathElementBasic():
	_IDENTIFIER = None
	pos: Vector2D
	relative: bool

	def serialize(self):
		return f"{self._IDENTIFIER if self.relative else self._IDENTIFIER.upper()} {self.pos.x} {self.pos.y}"

	def apply(self, pos):
		if self.relative:
			return pos + self.pos
		else:
			return self.pos

class SVGPathElementMove(SVGPathElementBasic):
	_IDENTIFIER = "m"

class SVGPathElementLine(SVGPathElementBasic):
	_IDENTIFIER = "l"

@dataclasses.dataclass
class SVGPathElementArc():
	_IDENTIFIER = "a"

	pos: Vector2D
	radius: Vector2D
	xrotation: float
	large_arc: bool
	sweep: bool
	relative: bool

	def serialize(self):
		return f"{self._IDENTIFIER if self.relative else self._IDENTIFIER.upper()} {self.radius.x} {self.radius.y} {self.xrotation} {1 if self.large_arc else 0} {1 if self.sweep else 0} {self.pos.x} {self.pos.y}"

	def apply(self, pos):
		if self.relative:
			return pos + self.pos
		else:
			return self.pos


@dataclasses.dataclass
class SVGPathElementBezier():
	_IDENTIFIER = "c"

	p1: Vector2D
	p2: Vector2D
	p3: Vector2D
	relative: bool

	def serialize(self):
		return f"{self._IDENTIFIER if self.relative else self._IDENTIFIER.upper()} {self.p1.x} {self.p1.y} {self.p2.x} {self.p2.y} {self.p3.x} {self.p3.y}"

	def apply(self, pos):
		if self.relative:
			return pos + self.p3
		else:
			return self.p3


@dataclasses.dataclass
class SVGPathElementHorizontal():
	_IDENTIFIER = "h"

	x: float
	relative: bool

	def serialize(self):
		return f"{self._IDENTIFIER if self.relative else self._IDENTIFIER.upper()} {self.x}"

	def apply(self, pos):
		if self.relative:
			return Vector2D(pos.x + self.x, pos.y)
		else:
			return Vector2D(self.x, pos.y)

@dataclasses.dataclass
class SVGPathElementVertical():
	_IDENTIFIER = "v"

	y: float
	relative: bool

	def serialize(self):
		return f"{self._IDENTIFIER if self.relative else self._IDENTIFIER.upper()} {self.y}"

	def apply(self, pos):
		if self.relative:
			return Vector2D(pos.x, pos.y + self.y)
		else:
			return Vector2D(pos.x, self.y)

class SVGPathParser():
	_CMD_BEGIN = re.compile(r"^[,\s]*(?P<cmd>[a-zA-Z])(?P<tail>.*)")
	_FLOAT = re.compile(r"^[,\s]*(?P<float>-?(\d*)?(\.\d*)?)(?P<tail>.*)")
	_INT = re.compile(r"^[,\s]*(?P<int>-?\d+)(?P<tail>.*)")

	def __init__(self):
		self._cmds = [ ]

	@property
	def cmds(self):
		return self._cmds

	def _parse_begin(self, text):
		rematch = self._CMD_BEGIN.fullmatch(text)
		rematch = rematch.groupdict()
		return (rematch["cmd"], rematch["tail"])

	def _parse_float(self, text):
		rematch = self._FLOAT.fullmatch(text)
		rematch = rematch.groupdict()
		return (float(rematch["float"]), rematch["tail"])

	def _parse_int(self, text):
		rematch = self._INT.fullmatch(text)
		rematch = rematch.groupdict()
		return (int(rematch["int"]), rematch["tail"])

	def _parse_bool(self, text):
		(intval, text) = self._parse_int(text)
		return (intval != 0, text)

	def _parse_pos(self, text):
		(x, text) = self._parse_float(text)
		(y, text) = self._parse_float(text)
		return (Vector2D(x, y), text)

	def parse(self, text):
		while len(text) > 0:
			(cmd, text) = self._parse_begin(text)
			if cmd in "mM":
				(pos, text) = self._parse_pos(text)
				self._cmds.append(SVGPathElementMove(pos, relative = cmd.islower()))
			elif cmd in "lL":
				(pos, text) = self._parse_pos(text)
				self._cmds.append(SVGPathElementLine(pos, relative = cmd.islower()))
			elif cmd in "hH":
				(x, text) = self._parse_float(text)
				self._cmds.append(SVGPathElementHorizontal(x, relative = cmd.islower()))
			elif cmd in "vV":
				(y, text) = self._parse_float(text)
				self._cmds.append(SVGPathElementVertical(y, relative = cmd.islower()))
			elif cmd in "aA":
				(radius, text) = self._parse_pos(text)
				(xrotation, text) = self._parse_float(text)
				(large_arc, text) = self._parse_bool(text)
				(sweep, text) = self._parse_bool(text)
				(pos, text) = self._parse_pos(text)
				self._cmds.append(SVGPathElementArc(radius = radius, xrotation = xrotation, large_arc = large_arc, sweep = sweep, pos = pos, relative = cmd.islower()))
			elif cmd in "cC":
				(p1, text) = self._parse_pos(text)
				(p2, text) = self._parse_pos(text)
				(p3, text) = self._parse_pos(text)
				self._cmds.append(SVGPathElementBezier(p1 = p1, p2 = p2, p3 = p3, relative = cmd.islower()))
			elif cmd in "zZ":
				self._cmds.append(SVGPathElementClose())
			else:
				raise NotImplementedError(cmd)


@SVGObject.register
class SVGPath(SVGObject, SVGStyleObject):
	_TAG_NAME = "path"

	def __init__(self, node):
		super().__init__(node)
		self._pos = Vector2D()
		for cmd in self.parsed:
			self._pos = cmd.apply(self._pos)

	@property
	def pos(self):
		return self._pos

	@property
	def parsed(self):
		if self.node.hasAttribute("d"):
			parser = SVGPathParser()
			parser.parse(self.node.getAttribute("d"))
			return parser.cmds
		else:
			return [ ]

	def clear(self, pos):
		self._pos = pos
		self.node.setAttribute("d", f"M {pos.x} {pos.y}")
		return self

	def __append_path(self, cmd):
		self._pos = cmd.apply(self._pos)
		self.node.setAttribute("d", f"{self.node.getAttribute('d')} {cmd.serialize()}")

	def horizontal(self, x, relative = False):
		return self.__append_path(SVGPathElementHorizontal(x = x, relative = relative))

	def vertical(self, y, relative = False):
		return self.__append_path(SVGPathElementVertical(y = y, relative = relative))

	def moveto(self, pos, relative = False):
		return self.__append_path(SVGPathElementMove(pos = pos, relative = relative))

	def lineto(self, pos, relative = False):
		return self.__append_path(SVGPathElementLine(pos = pos, relative = relative))

	def bezierto(self, p1, p2, p3, relative = False):
		return self.__append_path(SVGPathElementBezier(p1 = p1, p2 = p2, p3 = p3, relative = relative))

	def arcto(self, pos, radius, xrotation = 0, large_arc = True, sweep = True, relative = False):
		if isinstance(radius, float) or isinstance(radius, int):
			radius = Vector2D(radius, radius)
		return self.__append_path(SVGPathElementArc(pos = pos, radius = radius, xrotation = xrotation, large_arc = large_arc, sweep = sweep, relative = relative))

	def close(self):
		return self.__append_path(SVGPathElementClose())

	@contextlib.contextmanager
	def returnto(self):
		previous_pos = self.pos
		yield
		self.moveto(previous_pos)

	@classmethod
	def new(cls, pos):
		path = cls(cls._new_element())
		path._pos = pos
		path.node.setAttribute("d", f"M {pos.x} {pos.y}")
		path.style.default_path()
		return path



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2023 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
import logging
import functools
import subprocess
from .Exceptions import SVGValidationException

_log = logging.getLogger(__spec__.name)

class SVGValidatorErrorClass(enum.IntEnum):
	CheckDisabled = 0
	EmitWarning = 1
	ThrowException = 2

class SVGValidator():
	def __init__(self, check_missing_fonts: SVGValidatorErrorClass = SVGValidatorErrorClass.EmitWarning):
		self._check_missing_fonts = check_missing_fonts

	@functools.lru_cache(maxsize = 1000)
	def _have_font(self, font_family):
		return len(subprocess.check_output([ "fc-list", font_family ])) > 0

	def _check_font(self, font_name, missing_fonts):
		if font_name is None:
			return
		if font_name in missing_fonts:
			# Report only once.
			return
		if not self._have_font(font_name):
			missing_fonts.add(font_name)
			if self._check_missing_fonts == SVGValidatorErrorClass.EmitWarning:
				_log.warning("SVG document is referencing missing font: %s", font_name)
			elif self._check_missing_fonts == SVGValidatorErrorClass.ThrowException:
				raise SVGValidationException(f"SVG document is referencing missing font: {font_name}")

	def _validate_fonts(self, svg_document):
		missing_fonts = set()
		for svg_text in svg_document.walk("text"):
			self._check_font(svg_text.style.get("font-family", unstringify = True), missing_fonts)
			for tspan in svg_text.tspans:
				self._check_font(tspan.style.get("font-family", unstringify = True), missing_fonts)

	def validate(self, svg_document):
		if self._check_missing_fonts != SVGValidatorErrorClass.CheckDisabled:
			self._validate_fonts(svg_document)



#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

class SVGTransformation():
	_IDENTIFIER = None

	def __init__(self, svg_object):
		self._svg_object = svg_object

	@property
	def svg_object(self):
		return self._svg_object

	def apply(self):
		raise NotImplementedError(self.__class__.__name__)


class FormatTextTransformation(SVGTransformation):
	_IDENTIFIER = "format_text"

	def __init__(self, svg_object, template_vars):
		super().__init__(svg_object)
		self._template_vars = template_vars

	def apply(self):
		for tspan in self.svg_object.walk("tspan"):
			tspan.text = tspan.text.format(**self._template_vars)


class ChangeVisibilityTransformation(SVGTransformation):
	_IDENTIFIER = "visibility"

	def __init__(self, svg_object, visible):
		super().__init__(svg_object)
		self._visible = visible

	def apply(self):
		if self._visible:
			self.svg_object.style.show()
		else:
			self.svg_object.style.hide()



#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import pysvgedit
from .FriendlyArgumentParser import FriendlyArgumentParser

class ValidatorApp():
	def __init__(self, args):
		self._args = args

	def run(self):
		doc = pysvgedit.SVGDocument.readfile(self._args.infile_svg)
		validator = pysvgedit.SVGValidator()
		validator.validate(doc)

	@classmethod
	def main(cls):
		parser = FriendlyArgumentParser(description = "Validate an SVG file.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("infile_svg", help = "Input SVG file.")
		args = parser.parse_args(sys.argv[1:])

		app = cls(args)
		return app.run()



#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import enum
import logging
import functools
from .SVGTransformation import ChangeVisibilityTransformation
from .Exceptions import SVGInputFileException

_log = logging.getLogger(__spec__.name)

class SVGAnimationMode(enum.Enum):
	Compose = "compose"				# Add new layers on top
	ComposeAll = "compose-all"		# Add new layers on top, but consider all layers from the start
	Replace = "replace"				# New layers replace lower layers

class SVGLayerTag(enum.Enum):
	NoStop = "nostop"				# Do not emit this layer
	Protect = "protect"				# Do not remove this layer on 'reset'
	Reset = "reset"					# Remove all but protected layers

class SVGAnimation():
	def __init__(self, svg_document, animation_mode: SVGAnimationMode = SVGAnimationMode.Compose):
		self._svg_document = svg_document
		self._animation_mode = animation_mode

	@functools.cached_property
	def all_layers(self):
		return list(self._svg_document.get("g", constraint = lambda g: g.is_layer))

	@functools.cached_property
	def considered_layers(self):
		# Determine which layers we consider for the animation in the first place
		if self._animation_mode == SVGAnimationMode.Compose:
			considered_layers = self.visible_layers
		elif self._animation_mode in [ SVGAnimationMode.ComposeAll, SVGAnimationMode.Replace ]:
			considered_layers = self.all_layers
		else:
			raise NotImplementedError(self._animation_mode)
		return considered_layers

	@functools.cached_property
	def layer_tags(self):
		# Get all tags of considered layers
		return { layer.svgid: self._get_layer_tags(layer) for layer in self.considered_layers }

	@property
	def visible_layers(self):
		return [ layer for layer in self.all_layers if layer.style.is_visible ]

	def _get_layer_tags(self, layer):
		layer_tags = set()
		label = layer.label
		if ":" in label:
			(tag_string, _) = label.split(":", maxsplit = 1)
			for tag in tag_string.split(","):
				try:
					tag = SVGLayerTag(tag)
					layer_tags.add(tag)
				except ValueError as e:
					_log.warning("Unknown layer tag in layer %s (label %s): %s", layer.svgid, layer.label, tag)
		return layer_tags

	def _generate_layer_transformations(self):
		# Ensure all layers have unique layer IDs
		layer_ids = [layer.svgid for layer in self.considered_layers]
		if any(layer_id is None for layer_id in layer_ids):
			raise SVGInputFileException(f"{len(layer_ids)} layers found in SVG source, but some layers do not have an ID assigned.")
		if len(layer_ids) != len(set(layer_ids)):
			raise SVGInputFileException(f"{len(layer_ids)} layers found in SVG source, but some layers have duplicate layer IDs.")

		# Store commands to hide all layers first
		shown_unprotected_layers = [ ]
		svg_transforms = [ ChangeVisibilityTransformation(layer, visible = False) for layer in self.considered_layers ]

		# Then go through the layers one-by-one and render then as appropriate
		previous_layer = None
		for layer in self.considered_layers:
			tags = self.layer_tags[layer.svgid]
			_log.debug("Processing layer %s, tags %s", layer.svgid, tags)

			# Determine if slide is causing a reset, i.e. removal of below
			# layers which are unprotected
			if SVGLayerTag.Reset in tags:
				# Hide all below layers but those which are protected
				for below_layer in shown_unprotected_layers:
					svg_transforms.append(ChangeVisibilityTransformation(below_layer, visible = False))
				shown_unprotected_layers = [ ]

			# Then show the current layer
			svg_transforms.append(ChangeVisibilityTransformation(layer, visible = True))

			if SVGLayerTag.Protect not in tags:
				shown_unprotected_layers.append(layer)

			# If we're in a replacing mode, then hide the previous layer
			if (self._animation_mode == SVGAnimationMode.Replace) and (previous_layer is not None):
				svg_transforms.append(ChangeVisibilityTransformation(previous_layer, visible = False))

			# Emit it as a frame if it's not marked as "nostop"
			if SVGLayerTag.NoStop not in tags:
				yield svg_transforms
			previous_layer = layer

	def __iter__(self):
		for transformations in self._generate_layer_transformations():
			for transformation in transformations:
				transformation.apply()
			yield self._svg_document



#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from .Vector2D import Vector2D
from .SVGDefs import SVGDefs
from .SVGDocument import SVGDocument
from .SVGGroup import SVGGroup
from .SVGStyle import SVGStyle
from .SVGRect import SVGRect
from .SVGCircle import SVGCircle
from .SVGText import SVGText
from .SVGPath import SVGPath
from .SVGAnimation import SVGAnimation, SVGAnimationMode
from .SVGValidator import SVGValidator, SVGValidatorErrorClass
from .SVGTransformation import FormatTextTransformation, ChangeVisibilityTransformation
from .Exceptions import SVGException

VERSION = "0.0.3"



#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import xml.dom.minidom

class XMLTools():
	@classmethod
	def find_all_elements(cls, node, tagname):
		return (child for child in node.childNodes if (child.nodeType == child.ELEMENT_NODE) and (child.tagName == tagname))

	@classmethod
	def find_first_element(cls, node, tagname):
		return next(cls.find_all_elements(node, tagname))

	@classmethod
	def walk_elements(cls, node, tagname = None, constraint = None):
		if node.nodeType == node.ELEMENT_NODE:
			if ((constraint is None) or (constraint(node))) and ((tagname is None) or (node.tagName == tagname)):
				yield node
			for child in node.childNodes:
				yield from cls.walk_elements(child, tagname = tagname, constraint = constraint)

	@classmethod
	def default_get_attribute(cls, node, name, default_value = None):
		if node.hasAttribute(name):
			return node.getAttribute(name)
		else:
			return default_value

	@classmethod
	def new_element(cls, tagname):
		doc = xml.dom.minidom.Document()
		element = doc.createElement(tagname)
		return element

	@classmethod
	def try_remove_attribute(cls, node, name):
		if node.hasAttribute(name):
			node.removeAttribute(name)



#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import math

class Vector2D():
	def __init__(self, x = 0, y = 0):
		self._x = x
		self._y = y

	@property
	def x(self):
		return self._x

	@property
	def y(self):
		return self._y

	@property
	def length(self):
		return math.sqrt((self.x ** 2) + (self.y ** 2))

	@property
	def norm(self):
		return self / self.length

	@property
	def ortho(self):
		return Vector2D(self.y, self.x)

	def __matmul__(self, other):
		# Scalar product
		return (self.x * other.y) - (other.x * self.y)

	def __add__(self, vector):
		return Vector2D(self.x + vector.x, self.y + vector.y)

	def __mul__(self, scalar):
		return Vector2D(self.x * scalar, self.y * scalar)

	def __rmul__(self, scalar):
		return self * scalar

	def __truediv__(self, divisor):
		return Vector2D(self.x / divisor, self.y / divisor)

	def __repr__(self):
		return f"<{self.x:.1f}, {self.y:.1f}>"



#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import functools
import xml.dom.minidom
from .SVGObject import SVGObject, SVGWidthHeightObject
from .SVGDefs import SVGDefs
from .XMLTools import XMLTools

class SVGDocument(SVGObject, SVGWidthHeightObject):
	_TAG_NAME = "svg"
	_DEFAULT_WIDTH = 300
	_DEFAULT_HEIGHT = 150
	_NAMESPACES = {
		"inkscape": "http://www.inkscape.org/namespaces/inkscape",
		"sodipodi": "http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd",
		"svg": "http://www.w3.org/2000/svg",
		"rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
		"cc": "http://creativecommons.org/ns#",
		"dc": "http://purl.org/dc/elements/1.1/",
	}

	def __init__(self, svg_node):
		super().__init__(svg_node)
		svg_node.ownerDocument._pysvgedit = self
		self._used_ids = set(node.getAttribute("id") for node in XMLTools.walk_elements(svg_node) if node.hasAttribute("id"))

	def get_unused_id(self):
		ctr = len(self._used_ids) + 1
		while True:
			attempt_id = f"id{ctr}"
			if attempt_id not in self._used_ids:
				self._used_ids.add(attempt_id)
				return attempt_id

	@functools.cached_property
	def defs(self):
		try:
			return self.get_first("defs")
		except StopIteration:
			return self.add(SVGDefs.new())

	@classmethod
	def new(cls):
		doc = xml.dom.minidom.Document()
		root = doc.createElement("svg")
		root.setAttribute("xmlns", cls._NAMESPACES["svg"])
		for (nsname, nsvalue) in cls._NAMESPACES.items():
			root.setAttribute(f"xmlns:{nsname}", nsvalue)
		doc.appendChild(root)
		return cls(svg_node = root)

	@classmethod
	def frombytes(cls, bytes_data):
		doc = xml.dom.minidom.parseString(bytes_data)
		root = XMLTools.find_first_element(doc, "svg")
		return cls(root)

	@classmethod
	def read(cls, f):
		doc = xml.dom.minidom.parse(f)
		root = XMLTools.find_first_element(doc, "svg")
		return cls(root)

	@classmethod
	def readfile(cls, filename):
		with open(filename, "rb") as f:
			return cls.read(f)

	def asbytes(self):
		return self.node.ownerDocument.toxml(encoding = "utf-8")

	def write(self, f):
		self.node.ownerDocument.writexml(f)

	def writefile(self, filename):
		with open(filename, "w") as f:
			self.write(f)

	def get_element_by_id(self, element_id):
		constraint = lambda node: node.getAttribute("id") == element_id
		try:
			return next(XMLTools.walk_elements(self.node, constraint = constraint))
		except StopIteration:
			return None

	def __str__(self):
		return f"SVGDocument<{self.extents.x:.0f} x {self.extents.y:.0f}>"



#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from .SVGRect import SVGRect
from .SVGObject import SVGObject, SVGXYObject, SVGStyleObject
from .XMLTools import XMLTools

@SVGObject.register
class SVGTextSpan(SVGObject, SVGXYObject, SVGStyleObject):
	_TAG_NAME = "tspan"

	def __init__(self, tspan_node):
		super().__init__(tspan_node)
		try:
			self._text_node = next(child for child in tspan_node.childNodes if (child.nodeType == child.TEXT_NODE))
		except StopIteration:
			# No text node, create one (with empty text)
			self._text_node = tspan_node.appendChild(tspan_node.ownerDocument.createTextNode(""))

	@classmethod
	def new(cls, pos = None, text = ""):
		tspan_node = cls._new_element()
		tspan_node.appendChild(tspan_node.ownerDocument.createTextNode(text))

		svg_textspan = cls(tspan_node)
		if pos is not None:
			svg_textspan.pos = pos
		return svg_textspan

	@property
	def text(self):
		return self._text_node.wholeText

	@text.setter
	def text(self, value):
		# If text is replaced, we need to remove the X/Y coordinates
		XMLTools.try_remove_attribute(self.node, "x")
		XMLTools.try_remove_attribute(self.node, "y")
		return self._text_node.replaceWholeText(value)

	def __repr__(self):
		return f"tspan<{self.text}>"

@SVGObject.register
class SVGText(SVGObject, SVGStyleObject):
	_TAG_NAME = "text"

	@classmethod
	def new(cls, pos = None, rect_extents = None, text = ""):
		svg_text = cls(cls._new_element())
		svg_text.node.setAttribute("xml:space", "preserve")
		if (pos is not None) and (rect_extents is None):
			# Normal text with position
			svg_text.pos = pos
		elif (pos is not None) and (rect_extents is not None):
			# Text within rectangle (as definition)
			def post_add_hook(parent):
				rect = parent.svg_document.defs.add(SVGRect.new(pos = pos, extents = rect_extents))
				rect.style.clear()
				svg_text.style.shape_inside = rect
			svg_text.post_add_hook = post_add_hook

		svg_text.style.default_text()
		svg_text.node.appendChild(SVGTextSpan.new(pos = pos, text = text).node)
		return svg_text

	@property
	def tspans(self):
		return (SVGTextSpan(node) for node in XMLTools.find_all_elements(self.node, "tspan"))

	@property
	def tspan(self):
		return next(self.tspans)



#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from .SVGObject import SVGObject, SVGStyleObject
from .XMLTools import XMLTools

@SVGObject.register
class SVGGroup(SVGObject, SVGStyleObject):
	_TAG_NAME = "g"

	@property
	def is_layer(self):
		return self.node.getAttribute("inkscape:groupmode") == "layer"

	@is_layer.setter
	def is_layer(self, value: bool):
		if value:
			self.node.setAttribute("inkscape:groupmode", "layer")
		else:
			XMLTools.try_remove_attribute(self.node, "inkscape:groupmode")

	@classmethod
	def new(cls, is_layer = False):
		group = cls(cls._new_element())
		group.is_layer = is_layer
		return group



#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import json
import tempfile
import subprocess
import datetime
import mako.template
import pysvgedit
from .FriendlyArgumentParser import FriendlyArgumentParser

class HelperClass():
	pass

class MakoRendererApp():
	def __init__(self, args):
		self._args = args
		with open(self._args.datafile_json) as f:
			self._template_data = json.load(f)
		self._update_helper_vars()

	def _update_helper_vars(self):
		if self._args.functions != "none":
			h = HelperClass()
			self._template_data["h"] = h

			if self._args.functions == "default":
				h.now = datetime.datetime.now()

	def _write_svg(self):
		self._doc.writefile(self._args.outfile)

	def _write_pdf(self):
		with tempfile.NamedTemporaryFile(prefix = "pysvgedit_", suffix = ".svg", mode = "w") as f:
			self._doc.write(f)
			f.flush()
			subprocess.check_call([ "inkscape", "-o", self._args.outfile, f.name])

	def _replace_text(self):
		for tspan in self._doc.walk("tspan"):
			text = tspan.text
			if text == "":
				continue
			template = mako.template.Template(text, strict_undefined = True)
			try:
				rendered_text = template.render(**self._template_data)
			except Exception as e:
				if not self._args.ignore_errors:
					raise
				rendered_text = f"{e.__class__.__name__}: {e}"
				if self._args.verbose >= 1:
					print(f"Error rendering {text}: {rendered_text}")
			if text != rendered_text:
				if self._args.verbose >= 2:
					print(f"{text} -> {rendered_text}")
				tspan.text = rendered_text

	def _patch_style(self):
		for (style_id, style_update) in self._template_data.get("svg_style_patches", { }).items():
			node = self._doc.get_element_by_id(style_id)
			if node is None:
				print(f"Style update of '{style_id}' requested, but no such element found.")
				continue
			style = pysvgedit.SVGStyle.from_node(node, auto_sync = True)
			style.update(style_update)

	def run(self):
		self._doc = pysvgedit.SVGDocument.readfile(self._args.infile_svg)
		self._replace_text()
		if self._args.patch_style:
			self._patch_style()

		if self._args.outfile.endswith(".pdf"):
			self._write_pdf()
		else:
			self._write_svg()

	@classmethod
	def main(cls):
		parser = FriendlyArgumentParser(description = "Render text blocks in SVG files using Mako.")
		parser.add_argument("-i", "--ignore-errors", action = "store_true", help = "By default, rendering is refused if there are errors. This will continue rendering and replace tpspans with the encountered issues instead.")
		parser.add_argument("-p", "--patch-style", action = "store_true", help = "Interpret an 'svg_style_patches' dictionary as elements for which style should be updated.")
		parser.add_argument("-f", "--functions", choices = [ "none", "default" ], default = "default", help = "Special functions to supply using the 'h' variable. Can be one of %(choices)s, defaults to %(default)s.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("datafile_json", help = "JSON file that contains the data that will be used as template variables.")
		parser.add_argument("infile_svg", help = "Input SVG file.")
		parser.add_argument("outfile", help = "Output file. If the extension is '.pdf', will be directly rendered to PDF using Inkscape. Otherwise, SVG is produced.")
		args = parser.parse_args(sys.argv[1:])

		app = cls(args)
		return app.run()



#	pysvgedit - SVG manipulation toolkit
#	Copyright (C) 2023-2023 Johannes Bauer
#
#	This file is part of pysvgedit.
#
#	pysvgedit is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysvgedit is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pysvgedit; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from .Vector2D import Vector2D
from .SVGObject import SVGObject, SVGXYObject, SVGWidthHeightObject, SVGStyleObject

@SVGObject.register
class SVGRect(SVGObject, SVGXYObject, SVGWidthHeightObject, SVGStyleObject):
	_TAG_NAME = "rect"

	@property
	def p1(self):
		return self.pos

	@property
	def p2(self):
		return Vector2D(self.pos.x, self.pos.y + self.extents.y)

	@property
	def p3(self):
		return self.pos + self.extents

	@property
	def p4(self):
		return Vector2D(self.pos.x + self.extents.x, self.pos.y)

	@classmethod
	def new(cls, pos, extents):
		path = cls(cls._new_element())
		path.pos = pos
		path.extents = extents
		path.style.default_path()
		return path
